#include <stack>
#include <time.h>
#include <iomanip>

#include "Plane3.h"
#include "VspBspTree.h"
#include "Mesh.h"
#include "common.h"
#include "ViewCell.h"
#include "Environment.h"
#include "Polygon3.h"
#include "Ray.h"
#include "AxisAlignedBox3.h"
#include "Exporter.h"
#include "Plane3.h"
#include "ViewCellBsp.h"
#include "ViewCellsManager.h"
#include "Beam.h"

//-- static members

/** Evaluates split plane classification with respect to the plane's
	contribution for a minimum number of ray splits.
*/
const float VspBspTree::sLeastRaySplitsTable[] = {0, 0, 1, 1, 0};
/** Evaluates split plane classification with respect to the plane's
	contribution for balanced rays.
*/
const float VspBspTree::sBalancedRaysTable[] = {1, -1, 0, 0, 0};

int BspMergeCandidate::sMaxPvsSize = 0;

int VspBspTree::sFrontId = 0;
int VspBspTree::sBackId = 0;
int VspBspTree::sFrontAndBackId = 0;

float BspMergeCandidate::sOverallCost = 0;

/********************************************************************/
/*                  class VspBspTree implementation                 */
/********************************************************************/

VspBspTree::VspBspTree():
mRoot(NULL),
mPvsUseArea(true),
mCostNormalizer(Limits::Small),
mViewCellsManager(NULL),
mOutOfBoundsCell(NULL),
mShowInvalidSpace(false),
mStoreRays(false)
{
	bool randomize = false;
	environment->GetBoolValue("VspBspTree.Construction.randomize", randomize);
	if (randomize)
		Randomize(); // initialise random generator for heuristics

	//-- termination criteria for autopartition
	environment->GetIntValue("VspBspTree.Termination.maxDepth", mTermMaxDepth);
	environment->GetIntValue("VspBspTree.Termination.minPvs", mTermMinPvs);
	environment->GetIntValue("VspBspTree.Termination.minRays", mTermMinRays);
	environment->GetFloatValue("VspBspTree.Termination.minArea", mTermMinArea);
	environment->GetFloatValue("VspBspTree.Termination.maxRayContribution", mTermMaxRayContribution);
	environment->GetFloatValue("VspBspTree.Termination.minAccRayLenght", mTermMinAccRayLength);
	environment->GetFloatValue("VspBspTree.Termination.maxCostRatio", mTermMaxCostRatio);
	environment->GetIntValue("VspBspTree.Termination.missTolerance", mTermMissTolerance);

	//-- factors for bsp tree split plane heuristics
	environment->GetFloatValue("VspBspTree.Factor.balancedRays", mBalancedRaysFactor);
	environment->GetFloatValue("VspBspTree.Factor.pvs", mPvsFactor);
	environment->GetFloatValue("VspBspTree.Termination.ct_div_ci", mCtDivCi);

	//-- max cost ratio for early tree termination
	environment->GetFloatValue("VspBspTree.Termination.maxCostRatio", mTermMaxCostRatio);

	//-- partition criteria
	environment->GetIntValue("VspBspTree.maxPolyCandidates", mMaxPolyCandidates);
	environment->GetIntValue("VspBspTree.maxRayCandidates", mMaxRayCandidates);
	environment->GetIntValue("VspBspTree.splitPlaneStrategy", mSplitPlaneStrategy);

	environment->GetFloatValue("VspBspTree.Construction.epsilon", mEpsilon);
	environment->GetIntValue("VspBspTree.maxTests", mMaxTests);

	// maximum and minimum number of view cells
	environment->GetIntValue("VspBspTree.Termination.maxViewCells", mMaxViewCells);
	environment->GetIntValue("VspBspTree.PostProcess.minViewCells", mMergeMinViewCells);
	environment->GetFloatValue("VspBspTree.PostProcess.maxCostRatio", mMergeMaxCostRatio);

	environment->GetBoolValue("VspBspTree.splitUseOnlyDrivingAxis", mOnlyDrivingAxis);
	environment->GetBoolValue("VspBspTree.PostProcess.useRaysForMerge", mUseRaysForMerge);

	environment->GetFloatValue("ViewCells.maxPvsRatio", mMaxPvsRatio);

	//-- termination criteria for axis aligned split
	environment->GetFloatValue("VspBspTree.Termination.AxisAligned.maxRayContribution", 
		mTermMaxRayContriForAxisAligned);
	environment->GetIntValue("VspBspTree.Termination.AxisAligned.minRays",
		mTermMinRaysForAxisAligned);

	//environment->GetFloatValue("VspBspTree.maxTotalMemory", mMaxTotalMemory);
	environment->GetFloatValue("VspBspTree.maxStaticMemory", mMaxMemory);

	//-- debug output
	Debug << "******* VSP BSP options ******** " << endl;
    Debug << "max depth: " << mTermMaxDepth << endl;
	Debug << "min PVS: " << mTermMinPvs << endl;
	Debug << "min area: " << mTermMinArea << endl;
	Debug << "min rays: " << mTermMinRays << endl;
	Debug << "max ray contri: " << mTermMaxRayContribution << endl;
	//Debug << "VSP BSP mininam accumulated ray lenght: ", mTermMinAccRayLength) << endl;
	Debug << "max cost ratio: " << mTermMaxCostRatio << endl;
	Debug << "miss tolerance: " << mTermMissTolerance << endl;
	Debug << "max view cells: " << mMaxViewCells << endl;
	Debug << "max polygon candidates: " << mMaxPolyCandidates << endl;
	Debug << "max plane candidates: " << mMaxRayCandidates << endl;
	Debug << "randomize: " << randomize << endl;

	Debug << "Split plane strategy: ";
	if (mSplitPlaneStrategy & RANDOM_POLYGON)
	{
		Debug << "random polygon ";
	}
	if (mSplitPlaneStrategy & AXIS_ALIGNED)
	{
		Debug << "axis aligned ";
	}
	if (mSplitPlaneStrategy & LEAST_RAY_SPLITS)
	{
		mCostNormalizer += mLeastRaySplitsFactor;
		Debug << "least ray splits ";
	}
	if (mSplitPlaneStrategy & BALANCED_RAYS)
	{
		mCostNormalizer += mBalancedRaysFactor;
		Debug << "balanced rays ";
	}
	if (mSplitPlaneStrategy & PVS)
	{
		mCostNormalizer += mPvsFactor;
		Debug << "pvs";
	}


	mSplitCandidates = new vector<SortableEntry>;

	Debug << endl;
}

BspViewCell *VspBspTree::GetOutOfBoundsCell()
{
	return mOutOfBoundsCell;
}


BspViewCell *VspBspTree::GetOrCreateOutOfBoundsCell()
{
	if (!mOutOfBoundsCell)
	{
		mOutOfBoundsCell = new BspViewCell();
		mOutOfBoundsCell->SetId(-1);
	}
	return mOutOfBoundsCell;
}


const BspTreeStatistics &VspBspTree::GetStatistics() const
{
	return mStat;
}


VspBspTree::~VspBspTree()
{
	DEL_PTR(mRoot);
	DEL_PTR(mSplitCandidates);
}

int VspBspTree::AddMeshToPolygons(Mesh *mesh,
								  PolygonContainer &polys,
								  MeshInstance *parent)
{
	FaceContainer::const_iterator fi;

	// copy the face data to polygons
	for (fi = mesh->mFaces.begin(); fi != mesh->mFaces.end(); ++ fi)
	{
		Polygon3 *poly = new Polygon3((*fi), mesh);

		if (poly->Valid(mEpsilon))
		{
			poly->mParent = parent; // set parent intersectable
			polys.push_back(poly);
		}
		else
			DEL_PTR(poly);
	}
	return (int)mesh->mFaces.size();
}

int VspBspTree::AddToPolygonSoup(const ViewCellContainer &viewCells,
							  PolygonContainer &polys,
							  int maxObjects)
{
	int limit = (maxObjects > 0) ?
		Min((int)viewCells.size(), maxObjects) : (int)viewCells.size();

	int polysSize = 0;

	for (int i = 0; i < limit; ++ i)
	{
		if (viewCells[i]->GetMesh()) // copy the mesh data to polygons
		{
			mBox.Include(viewCells[i]->GetBox()); // add to BSP tree aabb
			polysSize +=
				AddMeshToPolygons(viewCells[i]->GetMesh(),
								  polys,
								  viewCells[i]);
		}
	}

	return polysSize;
}

int VspBspTree::AddToPolygonSoup(const ObjectContainer &objects,
								 PolygonContainer &polys,
								 int maxObjects)
{
	int limit = (maxObjects > 0) ?
		Min((int)objects.size(), maxObjects) : (int)objects.size();

	for (int i = 0; i < limit; ++i)
	{
		Intersectable *object = objects[i];//*it;
		Mesh *mesh = NULL;

		switch (object->Type()) // extract the meshes
		{
		case Intersectable::MESH_INSTANCE:
			mesh = dynamic_cast<MeshInstance *>(object)->GetMesh();
			break;
		case Intersectable::VIEW_CELL:
			mesh = dynamic_cast<ViewCell *>(object)->GetMesh();
			break;
			// TODO: handle transformed mesh instances
		default:
			Debug << "intersectable type not supported" << endl;
			break;
		}

        if (mesh) // copy the mesh data to polygons
		{
			mBox.Include(object->GetBox()); // add to BSP tree aabb
			AddMeshToPolygons(mesh, polys, NULL);
		}
	}

	return (int)polys.size();
}

void VspBspTree::Construct(const VssRayContainer &sampleRays,
						   AxisAlignedBox3 *forcedBoundingBox)
{
    mStat.nodes = 1;
	mBox.Initialize(); 	// initialise BSP tree bounding box

	if (forcedBoundingBox)
		mBox = *forcedBoundingBox;
	
	PolygonContainer polys;
	RayInfoContainer *rays = new RayInfoContainer();

	VssRayContainer::const_iterator rit, rit_end = sampleRays.end();

	long startTime = GetTime();

	cout << "Extracting polygons from rays ... ";

	Intersectable::NewMail();

	int numObj = 0;

	//-- extract polygons intersected by the rays
	for (rit = sampleRays.begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = *rit;

		if ((mBox.IsInside(ray->mTermination) || !forcedBoundingBox) &&
			ray->mTerminationObject && 
			!ray->mTerminationObject->Mailed())
		{
			ray->mTerminationObject->Mail();
			MeshInstance *obj = dynamic_cast<MeshInstance *>(ray->mTerminationObject);
			AddMeshToPolygons(obj->GetMesh(), polys, obj);
			++ numObj;
			//-- compute bounding box
			if (!forcedBoundingBox)
				mBox.Include(ray->mTermination);
		}

		if ((mBox.IsInside(ray->mOrigin) || !forcedBoundingBox) &&
			ray->mOriginObject && 
			!ray->mOriginObject->Mailed())
		{
			ray->mOriginObject->Mail();
			MeshInstance *obj = dynamic_cast<MeshInstance *>(ray->mOriginObject);
			AddMeshToPolygons(obj->GetMesh(), polys, obj);
			++ numObj;

			//-- compute bounding box
			if (!forcedBoundingBox)
				mBox.Include(ray->mOrigin);
		}
	}

	mMaxPvs = (int)(mMaxPvsRatio * (float)numObj);
	
	Debug << "maximal pvs (i.e., pvs still considered as valid) : " << mMaxPvs << endl;
	//-- store rays
	for (rit = sampleRays.begin(); rit != rit_end; ++ rit)
	{
		VssRay *ray = *rit;

		float minT, maxT;

		// TODO: not very efficient to implictly cast between rays types
		if (mBox.GetRaySegment(*ray, minT, maxT))
		{
			float len = ray->Length();

			if (!len)
				len = Limits::Small;

			rays->push_back(RayInfo(ray, minT / len, maxT / len));
		}
	}

	mTermMinArea *= mBox.SurfaceArea(); // normalize
	mStat.polys = (int)polys.size();

	cout << "finished" << endl;

	Debug << "\nPolygon extraction: " << (int)polys.size() << " polys extracted from "
		  << (int)sampleRays.size() << " rays in "
		  << TimeDiff(startTime, GetTime())*1e-3 << " secs" << endl << endl;

	Construct(polys, rays);

	// clean up polygons
	CLEAR_CONTAINER(polys);
}


// return memory usage in MB
float VspBspTree::GetMemUsage() const
{
	return
		(sizeof(VspBspTree) +
		 mStat.Leaves() * sizeof(BspLeaf) +
		 mStat.Interior() * sizeof(BspInterior) +
		 mStat.accumRays * sizeof(RayInfo)) / (1024.0f * 1024.0f);
}



void VspBspTree::Construct(const PolygonContainer &polys, RayInfoContainer *rays)
{
	VspBspTraversalStack tStack;

	mRoot = new BspLeaf();

	// constrruct root node geometry
	BspNodeGeometry *geom = new BspNodeGeometry();
	ConstructGeometry(mRoot, *geom);

	VspBspTraversalData tData(mRoot,
							  new PolygonContainer(polys),
							  0,
							  rays,
                              ComputePvsSize(*rays),
							  geom->GetArea(),
							  geom);

	tStack.push(tData);

	mStat.Start();
	cout << "Contructing vsp bsp tree ... \n";

	long startTime = GetTime();	
	long interTime = GetTime();	
	mOutOfMemory = false;

	int nleaves = 500;

	while (!tStack.empty())
	{
		tData = tStack.top();
	    tStack.pop();		

		if (0 && !mOutOfMemory)
		{
			float mem = GetMemUsage();

			if (mem > mMaxMemory)
			{
				mOutOfMemory = true;
				Debug << "memory limit reached: " << mem << endl;
			}
		}

			// subdivide leaf node
		BspNode *r = Subdivide(tStack, tData);

		if (r == mRoot)
			Debug << "VSP BSP tree construction time spent at root: "
				  << TimeDiff(startTime, GetTime())*1e-3 << "s" << endl;

		if (mStat.Leaves() >= nleaves)
		{
			nleaves += 500;
				
			cout << "leaves=" << mStat.Leaves() << endl;
			Debug << "needed "
			  << TimeDiff(interTime, GetTime())*1e-3 << " secs to create 500 leaves" << endl;
			interTime = GetTime();
		}
	}

	Debug << "Used Memory: " << GetMemUsage() << " MB" << endl << endl;
	cout << "finished\n";

	mStat.Stop();
}


bool VspBspTree::TerminationCriteriaMet(const VspBspTraversalData &data) const
{
	return
		(((int)data.mRays->size() <= mTermMinRays) ||
		 (data.mPvs <= mTermMinPvs)   ||
		 (data.mArea <= mTermMinArea) ||
		 (mStat.Leaves() >= mMaxViewCells) ||
		 (data.GetAvgRayContribution() > mTermMaxRayContribution) ||
		 (data.mDepth >= mTermMaxDepth));
}


BspNode *VspBspTree::Subdivide(VspBspTraversalStack &tStack,
							   VspBspTraversalData &tData)
{
	BspNode *newNode = tData.mNode;

	if (!mOutOfMemory && !TerminationCriteriaMet(tData))
	{
		PolygonContainer coincident;

		VspBspTraversalData tFrontData;
		VspBspTraversalData tBackData;

		// create new interior node and two leaf nodes
		// or return leaf as it is (if maxCostRatio missed)
		newNode = SubdivideNode(tData, tFrontData, tBackData, coincident);

		if (!newNode->IsLeaf()) //-- continue subdivision
		{
			// push the children on the stack
			tStack.push(tFrontData);
			tStack.push(tBackData);

			// delete old leaf node
			DEL_PTR(tData.mNode);
		}
	}

	//-- terminate traversal and create new view cell
	if (newNode->IsLeaf())
	{
		BspLeaf *leaf = dynamic_cast<BspLeaf *>(newNode);
		BspViewCell *viewCell;

		if (!CheckValid(tData))
		{
			leaf->SetTreeValid(false);
			PropagateUpValidity(leaf);
			// view cell for invalid view space
			viewCell = GetOrCreateOutOfBoundsCell();
			++ mStat.invalidLeaves;
		}
		else
		{	// create new view cell for this leaf			
			viewCell = new BspViewCell();
		}
		
		leaf->SetViewCell(viewCell);
	
		//-- update pvs
		int conSamp = 0, sampCon = 0;
		AddToPvs(leaf, *tData.mRays, conSamp, sampCon);

		mStat.contributingSamples += conSamp;
		mStat.sampleContributions += sampCon;

		//-- store additional info
		if (mStoreRays)
		{
			RayInfoContainer::const_iterator it, it_end = tData.mRays->end();
			for (it = tData.mRays->begin(); it != it_end; ++ it)
				leaf->mVssRays.push_back((*it).mRay);
		}

		viewCell->mLeaves.push_back(leaf);
		viewCell->SetArea(tData.mArea);
		leaf->mArea = tData.mArea;

		EvaluateLeafStats(tData);		
	}

	//-- cleanup
	tData.Clear();

	return newNode;
}




BspNode *VspBspTree::SubdivideNode(VspBspTraversalData &tData,
								   VspBspTraversalData &frontData,
								   VspBspTraversalData &backData,
								   PolygonContainer &coincident)
{
	BspLeaf *leaf = dynamic_cast<BspLeaf *>(tData.mNode);
	
	// select subdivision plane
	Plane3 splitPlane;
	
	int maxCostMisses = tData.mMaxCostMisses;

	const bool success = 
		SelectPlane(splitPlane, leaf, tData, frontData, backData);

	if (!success)
	{
		++ maxCostMisses;

		if (maxCostMisses > mTermMissTolerance)
		{
			// terminate branch because of max cost
			++ mStat.maxCostNodes;
            return leaf;
		}
	}

	mStat.nodes += 2;

	//-- subdivide further
	BspInterior *interior = new BspInterior(splitPlane);

#ifdef _DEBUG
	Debug << interior << endl;
#endif

	//-- the front and back traversal data is filled with the new values
	frontData.mDepth = tData.mDepth + 1;
	frontData.mPolygons = new PolygonContainer();
	frontData.mRays = new RayInfoContainer();
	
	backData.mDepth = tData.mDepth + 1;
	backData.mPolygons = new PolygonContainer();
	backData.mRays = new RayInfoContainer();
	
	// subdivide rays
	SplitRays(interior->GetPlane(),
			  *tData.mRays,
			  *frontData.mRays,
			  *backData.mRays);

	// subdivide polygons
	SplitPolygons(interior->GetPlane(),
				  *tData.mPolygons,
	              *frontData.mPolygons,
				  *backData.mPolygons,
				  coincident);


	// how often was max cost ratio missed in this branch?
	frontData.mMaxCostMisses = maxCostMisses;
	backData.mMaxCostMisses = maxCostMisses;

	// compute pvs
	frontData.mPvs = ComputePvsSize(*frontData.mRays);
	backData.mPvs = ComputePvsSize(*backData.mRays);

	// split front and back node geometry and compute area
	if (mPvsUseArea)
	{
		// if geometry was not already computed
		if (!frontData.mGeometry && !backData.mGeometry)
		{
			frontData.mGeometry = new BspNodeGeometry();
			backData.mGeometry = new BspNodeGeometry();

			tData.mGeometry->SplitGeometry(*frontData.mGeometry,
										   *backData.mGeometry,
										   interior->GetPlane(),
										   mBox,
										   mEpsilon);
		
			//frontData.mArea = frontData.mGeometry->GetArea();
			//backData.mArea = backData.mGeometry->GetArea();
			frontData.mArea = frontData.mGeometry->GetVolume();
			backData.mArea = backData.mGeometry->GetVolume();
		}
	}
	else
	{
		frontData.mArea = (float)frontData.mRays->size();
		backData.mArea = (float)backData.mRays->size();
	}

	//-- create front and back leaf

	BspInterior *parent = leaf->GetParent();

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, interior);
		interior->SetParent(parent);
	}
	else // new root
	{
		mRoot = interior;
	}

	// and setup child links
	interior->SetupChildLinks(new BspLeaf(interior), new BspLeaf(interior));

	frontData.mNode = interior->GetFront();
	backData.mNode = interior->GetBack();

	//DEL_PTR(leaf);
	return interior;
}


void VspBspTree::AddToPvs(BspLeaf *leaf,
						  const RayInfoContainer &rays,
						  int &sampleContributions,
						  int &contributingSamples)
{
	sampleContributions = 0;
	contributingSamples = 0;

    RayInfoContainer::const_iterator it, it_end = rays.end();

	ViewCell *vc = leaf->GetViewCell();

	// add contributions from samples to the PVS
	for (it = rays.begin(); it != it_end; ++ it)
	{
		int contribution = 0;
		VssRay *ray = (*it).mRay;

		if (ray->mTerminationObject)
			contribution += vc->GetPvs().AddSample(ray->mTerminationObject);

		if (ray->mOriginObject)
			contribution += vc->GetPvs().AddSample(ray->mOriginObject);

		if (contribution)
		{
			sampleContributions += contribution;
			++ contributingSamples;
		}
		//leaf->mVssRays.push_back(ray);
	}
}

void VspBspTree::SortSplitCandidates(const RayInfoContainer &rays, const int axis)
{
	mSplitCandidates->clear();

	int requestedSize = 2 * (int)(rays.size());
	// creates a sorted split candidates array
	if (mSplitCandidates->capacity() > 500000 &&
		requestedSize < (int)(mSplitCandidates->capacity()/10) )
	{
    	delete mSplitCandidates;
   		mSplitCandidates = new vector<SortableEntry>;
	}

	mSplitCandidates->reserve(requestedSize);

	// insert all queries
	for(RayInfoContainer::const_iterator ri = rays.begin();	ri < rays.end(); ++ ri)
	{
		bool positive = (*ri).mRay->HasPosDir(axis);
		mSplitCandidates->push_back(SortableEntry(positive ? SortableEntry::ERayMin : SortableEntry::ERayMax,
												  (*ri).ExtrapOrigin(axis), (*ri).mRay));

		mSplitCandidates->push_back(SortableEntry(positive ? SortableEntry::ERayMax : SortableEntry::ERayMin,
												  (*ri).ExtrapTermination(axis), (*ri).mRay));
	}

	stable_sort(mSplitCandidates->begin(), mSplitCandidates->end());
}

float VspBspTree::BestCostRatioHeuristics(const RayInfoContainer &rays,
										  const AxisAlignedBox3 &box,
										  const int pvsSize,
										  const int &axis,
                                          float &position)
{
	int raysBack;
	int raysFront;
	int pvsBack;
	int pvsFront;

	SortSplitCandidates(rays, axis);

	// go through the lists, count the number of objects left and right
	// and evaluate the following cost funcion:
	// C = ct_div_ci  + (ql*rl + qr*rr)/queries

	int rl = 0, rr = (int)rays.size();
	int pl = 0, pr = pvsSize;

	float minBox = box.Min(axis);
	float maxBox = box.Max(axis);
	float sizeBox = maxBox - minBox;

	float minBand = minBox + 0.1f*(maxBox - minBox);
	float maxBand = minBox + 0.9f*(maxBox - minBox);

	float sum = rr*sizeBox;
	float minSum = 1e20f;

	Intersectable::NewMail();

	// set all object as belonging to the fron pvs
	for(RayInfoContainer::const_iterator ri = rays.begin(); ri != rays.end(); ++ ri)
	{
		if ((*ri).mRay->IsActive())
		{
			Intersectable *object = (*ri).mRay->mTerminationObject;

			if (object)
			{
				if (!object->Mailed())
				{
					object->Mail();
					object->mCounter = 1;
				}
				else
					++ object->mCounter;
			}
		}
	}

	Intersectable::NewMail();

	for (vector<SortableEntry>::const_iterator ci = mSplitCandidates->begin();
		ci < mSplitCandidates->end(); ++ ci)
	{
		VssRay *ray;

		switch ((*ci).type)
		{
			case SortableEntry::ERayMin:
				{
					++ rl;
					ray = (VssRay *) (*ci).ray;

					Intersectable *object = ray->mTerminationObject;

					if (object && !object->Mailed())
					{
						object->Mail();
						++ pl;
					}
					break;
				}
			case SortableEntry::ERayMax:
				{
					-- rr;
					ray = (VssRay *) (*ci).ray;

					Intersectable *object = ray->mTerminationObject;

					if (object)
					{
						if (-- object->mCounter == 0)
							-- pr;
					}

					break;
				}
		}

		// Note: sufficient to compare size of bounding boxes of front and back side?
		if ((*ci).value > minBand && (*ci).value < maxBand)
		{
			sum = pl*((*ci).value - minBox) + pr*(maxBox - (*ci).value);

			//  cout<<"pos="<<(*ci).value<<"\t q=("<<ql<<","<<qr<<")\t r=("<<rl<<","<<rr<<")"<<endl;
			// cout<<"cost= "<<sum<<endl;

			if (sum < minSum)
			{
				minSum = sum;
				position = (*ci).value;

				raysBack = rl;
				raysFront = rr;

				pvsBack = pl;
				pvsFront = pr;

			}
		}
	}

	float oldCost = (float)pvsSize;
	float newCost = mCtDivCi + minSum / sizeBox;
	float ratio = newCost / oldCost;

	//Debug << "costRatio=" << ratio << " pos=" << position << " t=" << (position - minBox) / (maxBox - minBox)
	//     <<"\t q=(" << queriesBack << "," << queriesFront << ")\t r=(" << raysBack << "," << raysFront << ")" << endl;

	return ratio;
}


float VspBspTree::SelectAxisAlignedPlane(Plane3 &plane,
										 const VspBspTraversalData &tData,
										 int &axis,
										 BspNodeGeometry **frontGeom,
										 BspNodeGeometry **backGeom,
										 float &frontArea,
										 float &backArea,
										 bool useKdSplit)
{
	const bool useCostHeuristics = false;

	//-- regular split
	float nPosition[3];
	float nCostRatio[3];
	float nFrontArea[3];
	float nBackArea[3];

	BspNodeGeometry *nFrontGeom[3];
	BspNodeGeometry *nBackGeom[3];

	int bestAxis = -1;

	// create bounding box of node geometry
	AxisAlignedBox3 box;
	box.Initialize();
	//TODO: for kd split geometry already is box
	if (1 && mPvsUseArea)
	{
		PolygonContainer::const_iterator it, it_end = tData.mGeometry->mPolys.end();

		for(it = tData.mGeometry->mPolys.begin(); it < it_end; ++ it)
			box.Include(*(*it));
	}
	else
	{
		RayInfoContainer::const_iterator ri, ri_end = tData.mRays->end();

		for(ri = tData.mRays->begin(); ri < ri_end; ++ ri)
			box.Include((*ri).ExtrapTermination());
	}

	int sAxis = box.Size().DrivingAxis();

	for (axis = 0; axis < 3; ++ axis)
	{
		nFrontGeom[axis] = new BspNodeGeometry();
		nBackGeom[axis] = new BspNodeGeometry();

		if (!mOnlyDrivingAxis || (axis == sAxis))
		{
			if (!useCostHeuristics)
			{
				nPosition[axis] = (box.Min()[axis] + box.Max()[axis]) * 0.5f;
				Vector3 normal(0,0,0); normal[axis] = 1.0f;
				
				if (useKdSplit)
				{
					nCostRatio[axis] = EvalAxisAlignedSplitCost(tData,
																box,
																axis,
																nPosition[axis]);
					
					Vector3 pos;
					
					pos = box.Max(); pos[axis] = nPosition[axis];
					const AxisAlignedBox3 backBox(box.Min(), pos);
					backBox.ExtractPolys(nBackGeom[axis]->mPolys);
					
					pos = box.Min(); pos[axis] = nPosition[axis];
					const AxisAlignedBox3 frontBox(pos, box.Max());
					frontBox.ExtractPolys(nFrontGeom[axis]->mPolys);

				}
				else
				{
					nCostRatio[axis] = SplitPlaneCost(Plane3(normal, nPosition[axis]), 
													  tData, *nFrontGeom[axis], *nBackGeom[axis],
													  nFrontArea[axis], nBackArea[axis]);
				}
			}
			else
			{
				nCostRatio[axis] =
					BestCostRatioHeuristics(*tData.mRays,
										    box,
											tData.mPvs,
											axis,
											nPosition[axis]);
			}

			if (bestAxis == -1)
			{
				bestAxis = axis;
			}

			else if (nCostRatio[axis] < nCostRatio[bestAxis])
			{
				bestAxis = axis;
			}

		}
	}

	//-- assign values
	axis = bestAxis;
	frontArea = nFrontArea[bestAxis];
	backArea = nBackArea[bestAxis];

	// assign best split nodes geometry 
	*frontGeom = nFrontGeom[bestAxis];
	*backGeom = nBackGeom[bestAxis];

	// and delete other geometry
	delete nFrontGeom[(bestAxis + 1) % 3];
	delete nBackGeom[(bestAxis + 2) % 3];

	//-- split plane
    Vector3 normal(0,0,0); normal[bestAxis] = 1;
	plane = Plane3(normal, nPosition[bestAxis]);

	return nCostRatio[bestAxis];
}


float VspBspTree::EvalAxisAlignedSplitCost(const VspBspTraversalData &data,
										   const AxisAlignedBox3 &box,
										   const int axis,
										   const float &position) const
{
	int pvsFront = 0;
	int pvsBack = 0;
	int pvsTotal = 0;

	// create unique ids for pvs heuristics
	GenerateUniqueIdsForPvs();

	const int pvsSize = data.mPvs;

	// this is the main ray classification loop!
	for(RayInfoContainer::iterator ri = data.mRays->begin();
		ri != data.mRays->end(); ++ ri)
	{
		if (!(*ri).mRay->IsActive())
			continue;

		// determine the side of this ray with respect to the plane
		float t;
		int side = (*ri).ComputeRayIntersection(axis, position, t);
	
		AddObjToPvs((*ri).mRay->mTerminationObject, side, pvsFront, pvsBack, pvsTotal);
		AddObjToPvs((*ri).mRay->mOriginObject, side, pvsFront, pvsBack, pvsTotal);
	}

	//-- only one of these options should be one
	//-- pvs + probability heuristics
	float pBack, pFront, pOverall;

	if (1)
	{
		// box length substitute for probability
		const float minBox = box.Min(axis);
		const float maxBox = box.Max(axis);

		pBack = position - minBox;
		pFront = maxBox - position;
		pOverall = maxBox - minBox;
	}
	else //-- area substitute for probability
	{
		pOverall = box.SurfaceArea();

		const bool useMidSplit = true;
		//const bool useMidSplit = false;	
					
		//-- simplified computation for mid split
		const int axis2 = (axis + 1) % 3;
		const int axis3 = (axis + 2) % 3;

		const float faceArea = 
			(box.Max(axis2) - box.Min(axis2)) *
			(box.Max(axis3) - box.Min(axis3));

		pBack = pFront = pOverall * 0.5f + faceArea;
	}

	//Debug << axis << " " << pvsSize << " " << pvsBack << " " << pvsFront << endl;
	//Debug << pFront << " " << pBack << " " << pOverall << endl;

	// float sum = raysBack*(position - minBox) + raysFront*(maxBox - position);
	const float newCost = pvsBack * pBack + pvsFront * pFront;
	//  float oldCost = leaf->mRays.size();
	const float oldCost = (float)pvsSize * pOverall;

	return  (mCtDivCi + newCost) / oldCost;
}



bool VspBspTree::SelectPlane(Plane3 &bestPlane,
							 BspLeaf *leaf,
							 VspBspTraversalData &data,
							 VspBspTraversalData &frontData,
							 VspBspTraversalData &backData)
{
	// simplest strategy: just take next polygon
	if (mSplitPlaneStrategy & RANDOM_POLYGON)
	{
        if (!data.mPolygons->empty())
		{
			const int randIdx = 
				(int)RandomValue(0, (Real)((int)data.mPolygons->size() - 1));
			Polygon3 *nextPoly = (*data.mPolygons)[randIdx];

			bestPlane = nextPoly->GetSupportingPlane();
			return true;
		}
	}

	//-- use heuristics to find appropriate plane

	// intermediate plane
	Plane3 plane;
	float lowestCost = MAX_FLOAT;
	
	// decides if the first few splits should be only axisAligned
	const bool onlyAxisAligned  = 
		(mSplitPlaneStrategy & AXIS_ALIGNED) &&
		((int)data.mRays->size() > mTermMinRaysForAxisAligned) &&
		((int)data.GetAvgRayContribution() < mTermMaxRayContriForAxisAligned);
	
	const int limit = onlyAxisAligned ? 0 : 
		Min((int)data.mPolygons->size(), mMaxPolyCandidates);

	float candidateCost;

	int maxIdx = (int)data.mPolygons->size();

	for (int i = 0; i < limit; ++ i)
	{
		//-- assure that no index is taken twice
		const int candidateIdx = (int)RandomValue(0, (Real)(-- maxIdx));
		Polygon3 *poly = (*data.mPolygons)[candidateIdx];

		// swap candidate to the end to avoid testing same plane
		std::swap((*data.mPolygons)[maxIdx], (*data.mPolygons)[candidateIdx]);
		//Polygon3 *poly = (*data.mPolygons)[(int)RandomValue(0, (int)polys.size() - 1)];

		// evaluate current candidate
		BspNodeGeometry fGeom, bGeom;
		float fArea, bArea;
		plane = poly->GetSupportingPlane();
		candidateCost = SplitPlaneCost(plane, data, fGeom, bGeom, fArea, bArea);
		
		if (candidateCost < lowestCost)
		{
			bestPlane = plane;
			lowestCost = candidateCost;
		}
	}

	//-- evaluate axis aligned splits
	int axis;
	BspNodeGeometry *fGeom, *bGeom;
	float fArea, bArea;

	candidateCost = SelectAxisAlignedPlane(plane, data, axis,
										   &fGeom, &bGeom, 
										   fArea, bArea,
										   onlyAxisAligned);	 

	if (candidateCost < lowestCost)
	{	
		bestPlane = plane;
		lowestCost = candidateCost;

		// assign already computed values
		// we can do this because we always save the
		// computed values from the axis aligned splits
		frontData.mGeometry = fGeom;
		backData.mGeometry = bGeom;
		frontData.mArea = fArea;
		backData.mArea = bArea;

		//! error also computed if cost ratio is missed
		++ mStat.splits[axis];
	}
	else
	{
		DEL_PTR(fGeom);
		DEL_PTR(bGeom);
	}

#ifdef _DEBUG
	Debug << "plane lowest cost: " << lowestCost << endl;
#endif

	// cost ratio miss
	if (lowestCost > mTermMaxCostRatio)
		return false;

	return true;
}


Plane3 VspBspTree::ChooseCandidatePlane(const RayInfoContainer &rays) const
{
	const int candidateIdx = (int)RandomValue(0, (Real)((int)rays.size() - 1));

	const Vector3 minPt = rays[candidateIdx].ExtrapOrigin();
	const Vector3 maxPt = rays[candidateIdx].ExtrapTermination();

	const Vector3 pt = (maxPt + minPt) * 0.5;
	const Vector3 normal = Normalize(rays[candidateIdx].mRay->GetDir());

	return Plane3(normal, pt);
}


Plane3 VspBspTree::ChooseCandidatePlane2(const RayInfoContainer &rays) const
{
	Vector3 pt[3];

	int idx[3];
	int cmaxT = 0;
	int cminT = 0;
	bool chooseMin = false;

	for (int j = 0; j < 3; ++ j)
	{
		idx[j] = (int)RandomValue(0, (Real)((int)rays.size() * 2 - 1));

		if (idx[j] >= (int)rays.size())
		{
			idx[j] -= (int)rays.size();

			chooseMin = (cminT < 2);
		}
		else
			chooseMin = (cmaxT < 2);

		RayInfo rayInf = rays[idx[j]];
		pt[j] = chooseMin ? rayInf.ExtrapOrigin() : rayInf.ExtrapTermination();
	}

	return Plane3(pt[0], pt[1], pt[2]);
}

Plane3 VspBspTree::ChooseCandidatePlane3(const RayInfoContainer &rays) const
{
	Vector3 pt[3];

	int idx1 = (int)RandomValue(0, (Real)((int)rays.size() - 1));
	int idx2 = (int)RandomValue(0, (Real)((int)rays.size() - 1));

	// check if rays different
	if (idx1 == idx2)
		idx2 = (idx2 + 1) % (int)rays.size();

	const RayInfo ray1 = rays[idx1];
	const RayInfo ray2 = rays[idx2];

	// normal vector of the plane parallel to both lines
	const Vector3 norm = Normalize(CrossProd(ray1.mRay->GetDir(), ray2.mRay->GetDir()));

	// vector from line 1 to line 2
	const Vector3 vd = ray2.ExtrapOrigin() - ray1.ExtrapOrigin();

	// project vector on normal to get distance
	const float dist = DotProd(vd, norm);

	// point on plane lies halfway between the two planes
	const Vector3 planePt = ray1.ExtrapOrigin() + norm * dist * 0.5;

	return Plane3(norm, planePt);
}


inline void VspBspTree::GenerateUniqueIdsForPvs()
{
	Intersectable::NewMail(); sBackId = ViewCell::sMailId;
	Intersectable::NewMail(); sFrontId = ViewCell::sMailId;
	Intersectable::NewMail(); sFrontAndBackId = ViewCell::sMailId;
}


float VspBspTree::SplitPlaneCost(const Plane3 &candidatePlane,
								 const VspBspTraversalData &data,
								 BspNodeGeometry &geomFront,
								 BspNodeGeometry &geomBack,
								 float &areaFront,
								 float &areaBack) const
{
	float cost = 0;

	float sumBalancedRays = 0;
	float sumRaySplits = 0;

	int pvsFront = 0;
	int pvsBack = 0;

	// probability that view point lies in back / front node
	float pOverall = 0;
	float pFront = 0;
	float pBack = 0;

	int raysFront = 0;
	int raysBack = 0;
	int totalPvs = 0;

	int limit;
	bool useRand;

	// choose test rays randomly if too much
	if ((int)data.mRays->size() > mMaxTests)
	{
		useRand = true;
		limit = mMaxTests;
	}
	else
	{
		useRand = false;
		limit = (int)data.mRays->size();
	}
	
	for (int i = 0; i < limit; ++ i)
	{
		const int testIdx = useRand ? 
			(int)RandomValue(0, (Real)((int)data.mRays->size() - 1)) : i;
		RayInfo rayInf = (*data.mRays)[testIdx];

		float t;
		VssRay *ray = rayInf.mRay;
		const int cf = rayInf.ComputeRayIntersection(candidatePlane, t);

        if (cf >= 0)
			++ raysFront;
		if (cf <= 0)
			++ raysBack;

		if (mSplitPlaneStrategy & LEAST_RAY_SPLITS)
		{
			sumBalancedRays += cf;
		}

		if (mSplitPlaneStrategy & BALANCED_RAYS)
		{
			if (cf == 0)
				++ sumRaySplits;
		}

		if (mSplitPlaneStrategy & PVS)
		{
			// find front and back pvs for origing and termination object
			AddObjToPvs(ray->mTerminationObject, cf, pvsFront, pvsBack, totalPvs);
			AddObjToPvs(ray->mOriginObject, cf, pvsFront, pvsBack, totalPvs);
		}
	}

	const float raysSize = (float)data.mRays->size() + Limits::Small;
	if (mSplitPlaneStrategy & PVS)
	{
		// create unique ids for pvs heuristics
		GenerateUniqueIdsForPvs();

		if (mPvsUseArea) // use front and back cell areas to approximate volume
		{
			// construct child geometry with regard to the candidate split plane
			data.mGeometry->SplitGeometry(geomFront,
										  geomBack,
										  candidatePlane,
										  mBox,
										  mEpsilon);

			
			pFront = geomFront.GetVolume();
			pBack = geomBack.GetVolume();
			//areaFront = geomFront.GetArea();
			//areaBack = geomBack.GetArea();

			pOverall = data.mArea;

		}
		else // use number of rays to approximate volume
		{
			pOverall = (float)data.mRays->size();
			pFront = (float)raysFront;
			pBack = (float)raysBack;
		}
	}


	if (mSplitPlaneStrategy & LEAST_RAY_SPLITS)
		cost += mLeastRaySplitsFactor * sumRaySplits / raysSize;

	if (mSplitPlaneStrategy & BALANCED_RAYS)
		cost += mBalancedRaysFactor * fabs(sumBalancedRays) / raysSize;

	// pvs criterium
	if (mSplitPlaneStrategy & PVS)
	{
		const float oldCost = pOverall * (float)totalPvs + Limits::Small;
		cost += mPvsFactor * (pvsFront * pFront + pvsBack * pBack) / oldCost;
	}

#ifdef _DEBUG
  	Debug << "totalpvs: " << data.mPvs << " ptotal: " << pOverall
  		  << " frontpvs: " << pvsFront << " pFront: " << pFront
  		  << " backpvs: " << pvsBack << " pBack: " << pBack << endl << endl;
#endif

	// normalize cost by sum of linear factors
	return cost / (float)mCostNormalizer;
}


void VspBspTree::AddObjToPvs(Intersectable *obj,
							 const int cf,
							 int &frontPvs,
							 int &backPvs,
							 int &totalPvs) const
{
	if (!obj)
		return;
	
	if ((obj->mMailbox != sFrontId) &&
		(obj->mMailbox != sBackId) &&
		(obj->mMailbox != sFrontAndBackId))
	{
		++ totalPvs;
	}

	// TODO: does this really belong to no pvs?
	//if (cf == Ray::COINCIDENT) return;

	// object belongs to both PVS
	if (cf >= 0)
	{
		if ((obj->mMailbox != sFrontId) &&
			(obj->mMailbox != sFrontAndBackId))
		{
			++ frontPvs;
		
			if (obj->mMailbox == sBackId)
				obj->mMailbox = sFrontAndBackId;
			else
				obj->mMailbox = sFrontId;
		}
	}

	if (cf <= 0)
	{
		if ((obj->mMailbox != sBackId) &&
			(obj->mMailbox != sFrontAndBackId))
		{
			++ backPvs;
		
			if (obj->mMailbox == sFrontId)
				obj->mMailbox = sFrontAndBackId;
			else
				obj->mMailbox = sBackId;
		}
	}
}


void VspBspTree::CollectLeaves(vector<BspLeaf *> &leaves, 
							   const bool onlyUnmailed,
							   const int maxPvsSize) const
{
	stack<BspNode *> nodeStack;
	nodeStack.push(mRoot);

	while (!nodeStack.empty())
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			// test if this leaf is in valid view space
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);
			if (leaf->TreeValid() && 
				(!onlyUnmailed || !leaf->Mailed()) &&
				((maxPvsSize < 0) || (leaf->GetViewCell()->GetPvs().GetSize() <= maxPvsSize)))
			{
				leaves.push_back(leaf);
			}
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			nodeStack.push(interior->GetBack());
			nodeStack.push(interior->GetFront());
		}
	}
}


AxisAlignedBox3 VspBspTree::GetBoundingBox() const
{
	return mBox;
}


BspNode *VspBspTree::GetRoot() const
{
	return mRoot;
}


void VspBspTree::EvaluateLeafStats(const VspBspTraversalData &data)
{
	// the node became a leaf -> evaluate stats for leafs
	BspLeaf *leaf = dynamic_cast<BspLeaf *>(data.mNode);

	// store maximal and minimal depth
	if (data.mDepth > mStat.maxDepth)
		mStat.maxDepth = data.mDepth;

	if (data.mPvs > mStat.maxPvs)
		mStat.maxPvs = data.mPvs;
	
	if (data.mDepth < mStat.minDepth)
		mStat.minDepth = data.mDepth;

	if (data.mDepth >= mTermMaxDepth)
		++ mStat.maxDepthNodes;
	// accumulate rays to compute rays /  leaf
	mStat.accumRays += (int)data.mRays->size();

	if (data.mPvs < mTermMinPvs)
		++ mStat.minPvsNodes;

	if ((int)data.mRays->size() < mTermMinRays)
		++ mStat.minRaysNodes;

	if (data.GetAvgRayContribution() > mTermMaxRayContribution)
		++ mStat.maxRayContribNodes;

	if (data.mArea <= mTermMinArea)
		++ mStat.minAreaNodes;
	
	// accumulate depth to compute average depth
	mStat.accumDepth += data.mDepth;

#ifdef _DEBUG
	Debug << "BSP stats: "
		  << "Depth: " << data.mDepth << " (max: " << mTermMaxDepth << "), "
		  << "PVS: " << data.mPvs << " (min: " << mTermMinPvs << "), "
		  << "Area: " << data.mArea << " (min: " << mTermMinArea << "), "
		  << "#rays: " << (int)data.mRays->size() << " (max: " << mTermMinRays << "), "
		  << "#pvs: " << leaf->GetViewCell()->GetPvs().GetSize() << "=, "
		  << "#avg ray contrib (pvs): " << (float)data.mPvs / (float)data.mRays->size() << endl;
#endif
}

int VspBspTree::CastRay(Ray &ray)
{
	int hits = 0;

	stack<BspRayTraversalData> tStack;

	float maxt, mint;

	if (!mBox.GetRaySegment(ray, mint, maxt))
		return 0;

	Intersectable::NewMail();

	Vector3 entp = ray.Extrap(mint);
	Vector3 extp = ray.Extrap(maxt);

	BspNode *node = mRoot;
	BspNode *farChild = NULL;

	while (1)
	{
		if (!node->IsLeaf())
		{
			BspInterior *in = dynamic_cast<BspInterior *>(node);

			Plane3 splitPlane = in->GetPlane();
			const int entSide = splitPlane.Side(entp);
			const int extSide = splitPlane.Side(extp);

			if (entSide < 0)
			{
				node = in->GetBack();

				if(extSide <= 0) // plane does not split ray => no far child
					continue;

				farChild = in->GetFront(); // plane splits ray

			} else if (entSide > 0)
			{
				node = in->GetFront();

				if (extSide >= 0) // plane does not split ray => no far child
					continue;

				farChild = in->GetBack(); // plane splits ray
			}
			else // ray and plane are coincident
			{
				// WHAT TO DO IN THIS CASE ?
				//break;
				node = in->GetFront();
				continue;
			}

			// push data for far child
			tStack.push(BspRayTraversalData(farChild, extp, maxt));

			// find intersection of ray segment with plane
			float t;
			extp = splitPlane.FindIntersection(ray.GetLoc(), extp, &t);
			maxt *= t;

		} else // reached leaf => intersection with view cell
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			if (!leaf->GetViewCell()->Mailed())
			{
				//ray.bspIntersections.push_back(Ray::VspBspIntersection(maxt, leaf));
				leaf->GetViewCell()->Mail();
				++ hits;
			}

   			//-- fetch the next far child from the stack
			if (tStack.empty())
				break;

			entp = extp;
			mint = maxt; // NOTE: need this?

			if (ray.GetType() == Ray::LINE_SEGMENT && mint > 1.0f)
				break;

			BspRayTraversalData &s = tStack.top();

			node = s.mNode;
			extp = s.mExitPoint;
			maxt = s.mMaxT;

			tStack.pop();
		}
	}

	return hits;
}


void VspBspTree::CollectViewCells(ViewCellContainer &viewCells) const
{
	ViewCell::NewMail();
	CollectViewCells(mRoot, viewCells, true);
}


void VspBspTree::ValidateTree()
{
	stack<BspNode *> nodeStack;

	if (!mRoot)
		return;

	nodeStack.push(mRoot);
	
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			if (node->TreeValid())
			{
				BspViewCell *viewCell = dynamic_cast<BspLeaf *>(node)->GetViewCell();
			
				if (viewCell->GetPvs().GetSize() > mMaxPvs)
				{
					while (!viewCell->mLeaves.empty())
					{
						BspLeaf *l = viewCell->mLeaves.back();
						l->SetTreeValid(false);
						++ mStat.invalidLeaves;
						PropagateUpValidity(l);

						l->SetViewCell(GetOrCreateOutOfBoundsCell());
						viewCell->mLeaves.pop_back();
					}

					mOutOfBoundsCell->GetPvs().AddPvs(viewCell->GetPvs());

					DEL_PTR(viewCell);
				}
			}
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);
		
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}
}

void VspBspTree::CollectViewCells(BspNode *root, 
								  ViewCellContainer &viewCells,
								  bool onlyUnmailed) const
{
	stack<BspNode *> nodeStack;

	if (!root)
		return;

	nodeStack.push(root);
	
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();
		
		if (node->IsLeaf())
		{
			if (!mShowInvalidSpace && node->TreeValid())
			{
				ViewCell *viewCell = dynamic_cast<BspLeaf *>(node)->GetViewCell();
			
				if (!onlyUnmailed || !viewCell->Mailed()) 
				{
					viewCell->Mail();
					viewCells.push_back(viewCell);
				}
			}
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);
		
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}
}


float VspBspTree::AccumulatedRayLength(const RayInfoContainer &rays) const
{
	float len = 0;

	RayInfoContainer::const_iterator it, it_end = rays.end();

	for (it = rays.begin(); it != it_end; ++ it)
		len += (*it).SegmentLength();

	return len;
}


int VspBspTree::SplitRays(const Plane3 &plane,
						  RayInfoContainer &rays,
						  RayInfoContainer &frontRays,
						  RayInfoContainer &backRays)
{
	int splits = 0;

	while (!rays.empty())
	{
		RayInfo bRay = rays.back();
		rays.pop_back();

		VssRay *ray = bRay.mRay;
		float t;

		// get classification and receive new t
		const int cf = bRay.ComputeRayIntersection(plane, t);

		switch (cf)
		{
		case -1:
			backRays.push_back(bRay);
			break;
		case 1:
			frontRays.push_back(bRay);
			break;
		case 0:
			{
				//-- split ray
				//-- test if start point behind or in front of plane
				const int side = plane.Side(bRay.ExtrapOrigin());

				if (side <= 0)
				{
					backRays.push_back(RayInfo(ray, bRay.GetMinT(), t));
					frontRays.push_back(RayInfo(ray, t, bRay.GetMaxT()));
				}
				else
				{
					frontRays.push_back(RayInfo(ray, bRay.GetMinT(), t));
					backRays.push_back(RayInfo(ray, t, bRay.GetMaxT()));
				}
			}
			break;
		default:
			Debug << "Should not come here" << endl;
			break;
		}
	}

	return splits;
}


void VspBspTree::ExtractHalfSpaces(BspNode *n, vector<Plane3> &halfSpaces) const
{
	BspNode *lastNode;

	do
	{
		lastNode = n;

		// want to get planes defining geometry of this node => don't take
		// split plane of node itself
		n = n->GetParent();

		if (n)
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(n);
			Plane3 halfSpace = dynamic_cast<BspInterior *>(interior)->GetPlane();

            if (interior->GetFront() != lastNode)
				halfSpace.ReverseOrientation();

			halfSpaces.push_back(halfSpace);
		}
	}
	while (n);
}


void VspBspTree::ConstructGeometry(BspNode *n,
								   BspNodeGeometry &geom) const
{
	vector<Plane3> halfSpaces;
	ExtractHalfSpaces(n, halfSpaces);

	PolygonContainer candidatePolys;

	// bounded planes are added to the polygons (reverse polygons
	// as they have to be outfacing
	for (int i = 0; i < (int)halfSpaces.size(); ++ i)
	{
		Polygon3 *p = GetBoundingBox().CrossSection(halfSpaces[i]);

		if (p->Valid(mEpsilon))
		{
			candidatePolys.push_back(p->CreateReversePolygon());
			DEL_PTR(p);
		}
	}

	// add faces of bounding box (also could be faces of the cell)
	for (int i = 0; i < 6; ++ i)
	{
		VertexContainer vertices;

		for (int j = 0; j < 4; ++ j)
			vertices.push_back(mBox.GetFace(i).mVertices[j]);

		candidatePolys.push_back(new Polygon3(vertices));
	}

	for (int i = 0; i < (int)candidatePolys.size(); ++ i)
	{
		// polygon is split by all other planes
		for (int j = 0; (j < (int)halfSpaces.size()) && candidatePolys[i]; ++ j)
		{
			if (i == j) // polygon and plane are coincident
				continue;

			VertexContainer splitPts;
			Polygon3 *frontPoly, *backPoly;

			const int cf =
				candidatePolys[i]->ClassifyPlane(halfSpaces[j],
												 mEpsilon);

			switch (cf)
			{
				case Polygon3::SPLIT:
					frontPoly = new Polygon3();
					backPoly = new Polygon3();

					candidatePolys[i]->Split(halfSpaces[j],
											 *frontPoly,
											 *backPoly,
											 mEpsilon);

					DEL_PTR(candidatePolys[i]);

					if (frontPoly->Valid(mEpsilon))
						candidatePolys[i] = frontPoly;
					else
						DEL_PTR(frontPoly);

					DEL_PTR(backPoly);
					break;
				case Polygon3::BACK_SIDE:
					DEL_PTR(candidatePolys[i]);
					break;
				// just take polygon as it is
				case Polygon3::FRONT_SIDE:
				case Polygon3::COINCIDENT:
				default:
					break;
			}
		}

		if (candidatePolys[i])
			geom.mPolys.push_back(candidatePolys[i]);
	}
}


void VspBspTree::ConstructGeometry(BspViewCell *vc, 
								   BspNodeGeometry &vcGeom) const
{
	vector<BspLeaf *> leaves = vc->mLeaves;
	vector<BspLeaf *>::const_iterator it, it_end = leaves.end();

	for (it = leaves.begin(); it != it_end; ++ it)
		ConstructGeometry(*it, vcGeom);
}


int VspBspTree::FindNeighbors(BspNode *n, vector<BspLeaf *> &neighbors,
						   const bool onlyUnmailed) const
{
	BspNodeGeometry geom;
	ConstructGeometry(n, geom);

	stack<BspNode *> nodeStack;
	nodeStack.push(mRoot);

	// split planes from the root to this node
	// needed to verify that we found neighbor leaf
	vector<Plane3> halfSpaces;
	ExtractHalfSpaces(n, halfSpaces);

	while (!nodeStack.empty())
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())

		{	// test if this leaf is in valid view space
            if (node->TreeValid() &&
				node != n && 
				(!onlyUnmailed || !node->Mailed()))
			{
				// test all planes of current node if candidate really
				// is neighbour
				BspNodeGeometry neighborCandidate;
				ConstructGeometry(node, neighborCandidate);

				bool isAdjacent = true;
				for (int i = 0; (i < halfSpaces.size()) && isAdjacent; ++ i)
				{
					const int cf =
						Polygon3::ClassifyPlane(neighborCandidate.mPolys,
												halfSpaces[i],
												mEpsilon);

					if (cf == Polygon3::BACK_SIDE)
						isAdjacent = false;
				}
				// neighbor was found
				if (isAdjacent)
					neighbors.push_back(dynamic_cast<BspLeaf *>(node));
			}
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			const int cf = Polygon3::ClassifyPlane(geom.mPolys,
												   interior->GetPlane(),
												   mEpsilon);

			if (cf == Polygon3::FRONT_SIDE)
				nodeStack.push(interior->GetFront());
			else
				if (cf == Polygon3::BACK_SIDE)
					nodeStack.push(interior->GetBack());
				else
				{
					// random decision
					nodeStack.push(interior->GetBack());
					nodeStack.push(interior->GetFront());
				}
		}
	}

	return (int)neighbors.size();
}


BspLeaf *VspBspTree::GetRandomLeaf(const Plane3 &halfspace)
{
    stack<BspNode *> nodeStack;
  	nodeStack.push(mRoot);

	int mask = rand();

	while (!nodeStack.empty())
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())
		{
			return dynamic_cast<BspLeaf *>(node);
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);
			BspNode *next;
			BspNodeGeometry geom;

			// todo: not very efficient: constructs full cell everytime
			ConstructGeometry(interior, geom);

			const int cf = 
				Polygon3::ClassifyPlane(geom.mPolys, halfspace, mEpsilon);

			if (cf == Polygon3::BACK_SIDE)
				next = interior->GetFront();
			else
				if (cf == Polygon3::FRONT_SIDE)
					next = interior->GetFront();
			else
			{
				// random decision
				if (mask & 1)
					next = interior->GetBack();
				else
					next = interior->GetFront();
				mask = mask >> 1;
			}

			nodeStack.push(next);
		}
	}

	return NULL;
}

BspLeaf *VspBspTree::GetRandomLeaf(const bool onlyUnmailed)
{
	stack<BspNode *> nodeStack;

	nodeStack.push(mRoot);

	int mask = rand();

	while (!nodeStack.empty())
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())
		{
			if ( (!onlyUnmailed || !node->Mailed()) )
				return dynamic_cast<BspLeaf *>(node);
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// random decision
			if (mask & 1)
				nodeStack.push(interior->GetBack());
			else
				nodeStack.push(interior->GetFront());

			mask = mask >> 1;
		}
	}

	return NULL;
}

int VspBspTree::ComputePvsSize(const RayInfoContainer &rays) const
{
	int pvsSize = 0;

	RayInfoContainer::const_iterator rit, rit_end = rays.end();

	Intersectable::NewMail();

	for (rit = rays.begin(); rit != rays.end(); ++ rit)
	{
		VssRay *ray = (*rit).mRay;

		if (ray->mOriginObject)
		{
			if (!ray->mOriginObject->Mailed())
			{
				ray->mOriginObject->Mail();
				++ pvsSize;
			}
		}
		if (ray->mTerminationObject)
		{
			if (!ray->mTerminationObject->Mailed())
			{
				ray->mTerminationObject->Mail();
				++ pvsSize;
			}
		}
	}

	return pvsSize;
}

float VspBspTree::GetEpsilon() const
{
	return mEpsilon;
}


int VspBspTree::SplitPolygons(const Plane3 &plane,
							  PolygonContainer &polys,
							  PolygonContainer &frontPolys,
							  PolygonContainer &backPolys,
							  PolygonContainer &coincident) const
{
	int splits = 0;

	while (!polys.empty())
	{
		Polygon3 *poly = polys.back();
		polys.pop_back();

		// classify polygon
		const int cf = poly->ClassifyPlane(plane, mEpsilon);

		switch (cf)
		{
			case Polygon3::COINCIDENT:
				coincident.push_back(poly);
				break;
			case Polygon3::FRONT_SIDE:
				frontPolys.push_back(poly);
				break;
			case Polygon3::BACK_SIDE:
				backPolys.push_back(poly);
				break;
			case Polygon3::SPLIT:
				backPolys.push_back(poly);
				frontPolys.push_back(poly);
				++ splits;
				break;
			default:
                Debug << "SHOULD NEVER COME HERE\n";
				break;
		}
	}

	return splits;
}


int VspBspTree::CastLineSegment(const Vector3 &origin,
								const Vector3 &termination,
								vector<ViewCell *> &viewcells)
{
	int hits = 0;
	stack<BspRayTraversalData> tStack;

	float mint = 0.0f, maxt = 1.0f;

	Intersectable::NewMail();

	Vector3 entp = origin;
	Vector3 extp = termination;

	BspNode *node = mRoot;
	BspNode *farChild = NULL;

	float t;
	while (1)
	{
		if (!node->IsLeaf())
		{
			BspInterior *in = dynamic_cast<BspInterior *>(node);

			Plane3 splitPlane = in->GetPlane();
			
			const int entSide = splitPlane.Side(entp);
			const int extSide = splitPlane.Side(extp);

			if (entSide < 0) 
			{
				node = in->GetBack();
				// plane does not split ray => no far child
				if (extSide <= 0) 
					continue;

				farChild = in->GetFront(); // plane splits ray
			} 
			else if (entSide > 0)
			{
				node = in->GetFront();

				if (extSide >= 0) // plane does not split ray => no far child
					continue;

				farChild = in->GetBack(); // plane splits ray
			}
			else // ray end point on plane
			{	// NOTE: what to do if ray is coincident with plane?
				if (extSide < 0)
					node = in->GetBack();
				else 
					node = in->GetFront();
								
				continue; // no far child
			}

			// push data for far child
			tStack.push(BspRayTraversalData(farChild, extp));

			// find intersection of ray segment with plane
			extp = splitPlane.FindIntersection(origin, extp, &t);
		} 
		else
		{
			// reached leaf => intersection with view cell
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			if (!leaf->GetViewCell()->Mailed())
			{
				viewcells.push_back(leaf->GetViewCell());
				leaf->GetViewCell()->Mail();
				++ hits;
			}

			//-- fetch the next far child from the stack
			if (tStack.empty())
				break;

			entp = extp;
			
			BspRayTraversalData &s = tStack.top();

			node = s.mNode;
			extp = s.mExitPoint;

			tStack.pop();
		}
	}

	return hits;
}

int VspBspTree::TreeDistance(BspNode *n1, BspNode *n2) const
{
	std::deque<BspNode *> path1;
	BspNode *p1 = n1;

	// create path from node 1 to root
	while (p1)
	{
		if (p1 == n2) // second node on path
			return (int)path1.size();

		path1.push_front(p1);
		p1 = p1->GetParent();
	}

	int depth = n2->GetDepth();
	int d = depth;

	BspNode *p2 = n2;

	// compare with same depth
	while (1)
	{
		if ((d < (int)path1.size()) && (p2 == path1[d]))
			return (depth - d) + ((int)path1.size() - 1 - d);

		-- d;
		p2 = p2->GetParent();
	}

	return 0; // never come here
}

BspNode *VspBspTree::CollapseTree(BspNode *node, int &collapsed)
{
	if (node->IsLeaf())
		return node;

	BspInterior *interior = dynamic_cast<BspInterior *>(node);

	BspNode *front = CollapseTree(interior->GetFront(), collapsed);
	BspNode *back = CollapseTree(interior->GetBack(), collapsed);

	if (front->IsLeaf() && back->IsLeaf())
	{
		BspLeaf *frontLeaf = dynamic_cast<BspLeaf *>(front);
		BspLeaf *backLeaf = dynamic_cast<BspLeaf *>(back);

		//-- collapse tree
		if (frontLeaf->GetViewCell() == backLeaf->GetViewCell())
		{
			BspViewCell *vc = frontLeaf->GetViewCell();

			BspLeaf *leaf = new BspLeaf(interior->GetParent(), vc);
			leaf->SetTreeValid(frontLeaf->TreeValid());

			// replace a link from node's parent
			if (leaf->GetParent())
				leaf->GetParent()->ReplaceChildLink(node, leaf);
			else
				mRoot = leaf;

			++ collapsed;
			delete interior;

			return leaf;
		}
	}

	return node;
}


int VspBspTree::CollapseTree()
{
	int collapsed = 0;
	
	(void)CollapseTree(mRoot, collapsed);

	// revalidate leaves
	RepairViewCellsLeafLists();

	return collapsed;
}


void VspBspTree::RepairViewCellsLeafLists()
{
	// list not valid anymore => clear
	stack<BspNode *> nodeStack;
	nodeStack.push(mRoot);

	ViewCell::NewMail();

	while (!nodeStack.empty())
	{
		BspNode *node = nodeStack.top();
		nodeStack.pop();

		if (node->IsLeaf())
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			BspViewCell *viewCell = leaf->GetViewCell();

			if (!viewCell->Mailed())
			{
				viewCell->mLeaves.clear();
				viewCell->Mail();
			}

			viewCell->mLeaves.push_back(leaf);
		}
		else
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}
}


typedef pair<BspNode *, BspNodeGeometry *> bspNodePair;


int VspBspTree::CastBeam(Beam &beam)
{
    stack<bspNodePair> nodeStack;
	BspNodeGeometry *rgeom = new BspNodeGeometry();
	ConstructGeometry(mRoot, *rgeom);

	nodeStack.push(bspNodePair(mRoot, rgeom));
  
	ViewCell::NewMail();

	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top().first;
		BspNodeGeometry *geom = nodeStack.top().second;
		nodeStack.pop();
		
		AxisAlignedBox3 box;
		box.Initialize();
		geom->IncludeInBox(box);

		const int side = beam.ComputeIntersection(box);
		
		switch (side) 
		{
		case -1:
			CollectViewCells(node, beam.mViewCells, true);
			break;
		case 0:
			
			if (node->IsLeaf())
			{
				BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);
			
				if (!leaf->GetViewCell()->Mailed() && leaf->TreeValid())
				{
					leaf->GetViewCell()->Mail();
					beam.mViewCells.push_back(leaf->GetViewCell());
				}
			}
			else 
			{
				BspInterior *interior = dynamic_cast<BspInterior *>(node);
			
				BspNode *first = interior->GetFront();
				BspNode *second = interior->GetBack();
            
				BspNodeGeometry *firstGeom = new BspNodeGeometry();
				BspNodeGeometry *secondGeom = new BspNodeGeometry();

				geom->SplitGeometry(*firstGeom,
									*secondGeom,
									interior->GetPlane(),
									mBox,
									mEpsilon);

				// decide on the order of the nodes
				if (DotProd(beam.mPlanes[0].mNormal, 
					interior->GetPlane().mNormal) > 0)
				{
					swap(first, second);
					swap(firstGeom, secondGeom);
				}

				nodeStack.push(bspNodePair(first, firstGeom));
				nodeStack.push(bspNodePair(second, secondGeom));
			}
			
			break;
		default:
			// default: cull
			break;
		}
		
		DEL_PTR(geom);
		
	}

	return (int)beam.mViewCells.size();
}


bool VspBspTree::MergeViewCells(BspLeaf *l1, BspLeaf *l2) const
{
	//-- change pointer to view cells of all leaves associated
	//-- with the previous view cells
	BspViewCell *fVc = l1->GetViewCell();
	BspViewCell *bVc = l2->GetViewCell();

	BspViewCell *vc = dynamic_cast<BspViewCell *>(
		mViewCellsManager->MergeViewCells(*fVc, *bVc));

	// if merge was unsuccessful
	if (!vc) return false;

	// set new size of view cell
	vc->SetArea(fVc->GetArea() + bVc->GetArea());

	vector<BspLeaf *> fLeaves = fVc->mLeaves;
	vector<BspLeaf *> bLeaves = bVc->mLeaves;

	vector<BspLeaf *>::const_iterator it;

	//-- change view cells of all the other leaves the view cell belongs to
	for (it = fLeaves.begin(); it != fLeaves.end(); ++ it)
	{
		(*it)->SetViewCell(vc);
		vc->mLeaves.push_back(*it);
	}

	for (it = bLeaves.begin(); it != bLeaves.end(); ++ it)
	{
		(*it)->SetViewCell(vc);
		vc->mLeaves.push_back(*it);
	}

	vc->Mail();

	// clean up old view cells
	DEL_PTR(fVc);
	DEL_PTR(bVc);

	return true;
}


void VspBspTree::SetViewCellsManager(ViewCellsManager *vcm)
{
	mViewCellsManager = vcm;
}


int VspBspTree::CollectMergeCandidates(const vector<BspLeaf *> leaves)
{
	BspLeaf::NewMail();
	
	vector<BspLeaf *>::const_iterator it, it_end = leaves.end();

	int candidates = 0;

	// find merge candidates and push them into queue
	for (it = leaves.begin(); it != it_end; ++ it)
	{
		BspLeaf *leaf = *it;
		
		/// create leaf pvs (needed for post processing
		leaf->mPvs = new ObjectPvs(leaf->GetViewCell()->GetPvs());

		BspMergeCandidate::sOverallCost += 
			leaf->mArea * leaf->mPvs->GetSize();

		// the same leaves must not be part of two merge candidates
		leaf->Mail();
		vector<BspLeaf *> neighbors;
		FindNeighbors(leaf, neighbors, true);

		vector<BspLeaf *>::const_iterator nit, nit_end = neighbors.end();

		// TODO: test if at least one ray goes from one leaf to the other
		for (nit = neighbors.begin(); nit != nit_end; ++ nit)
		{
			if ((*nit)->GetViewCell() != leaf->GetViewCell())
			{
				mMergeQueue.push(BspMergeCandidate(leaf, *nit));
				++ candidates;
			}
		}
	}

	Debug << "found " << candidates << " new merge candidates" << endl;

	return (int)leaves.size();
}


int VspBspTree::CollectMergeCandidates(const VssRayContainer &rays)
{
	vector<BspRay *> bspRays;

	long startTime = GetTime();
	ConstructBspRays(bspRays, rays);
	Debug << (int)bspRays.size() << " bsp rays constructed in " 
		  << TimeDiff(startTime, GetTime()) * 1e-3f << " secs" << endl;

	map<BspLeaf *, vector<BspLeaf*> > neighborMap;
	vector<BspIntersection>::const_iterator iit;

	int numLeaves = 0;
	
	BspLeaf::NewMail();

	for (int i = 0; i < (int)bspRays.size(); ++ i)
	{  
		BspRay *ray = bspRays[i];
	  
		// traverse leaves stored in the rays and compare and 
		// merge consecutive leaves (i.e., the neighbors in the tree)
		if (ray->intersections.size() < 2)
			continue;
	  
		iit = ray->intersections.begin();
		BspLeaf *leaf = (*(iit ++)).mLeaf;
		
		// create leaf pvs (needed for post processing)
		if (!leaf->mPvs)
		{
			leaf->mPvs = 
				new ObjectPvs(leaf->GetViewCell()->GetPvs());

			BspMergeCandidate::sOverallCost += 
				leaf->mArea * leaf->mPvs->GetSize();
			
			++ numLeaves;
		}
		
		// traverse intersections 
		// consecutive leaves are neighbors => add them to queue
		for (; iit != ray->intersections.end(); ++ iit)
		{
			// next pair
			BspLeaf *prevLeaf = leaf;
            leaf = (*iit).mLeaf;

			// view space not valid or same view cell
			if (!leaf->TreeValid() || !prevLeaf->TreeValid() ||
				(leaf->GetViewCell() == prevLeaf->GetViewCell()))
				continue;

            // create leaf pvs (needed for post processing)
			if (!leaf->mPvs)
			{
				leaf->mPvs = 
					new ObjectPvs(leaf->GetViewCell()->GetPvs());
				
				BspMergeCandidate::sOverallCost += 
					leaf->mArea * leaf->mPvs->GetSize();

				++ numLeaves;
			}
		
			vector<BspLeaf *> &neighbors = neighborMap[leaf];
			
			bool found = false;

			// both leaves inserted in queue already =>
			// look if double pair already exists
			if (leaf->Mailed() && prevLeaf->Mailed())
			{
				vector<BspLeaf *>::const_iterator it, it_end = neighbors.end();
				
                for (it = neighbors.begin(); !found && (it != it_end); ++ it)
					if (*it == prevLeaf)
						found = true; // already in queue
			}
			
			if (!found)
			{
				// this pair is not in map already
				// => insert into the neighbor map and the queue
				neighbors.push_back(prevLeaf);
				neighborMap[prevLeaf].push_back(leaf);

				leaf->Mail();
				prevLeaf->Mail();

				mMergeQueue.push(BspMergeCandidate(leaf, prevLeaf));
			}
        }
	}
	Debug << "neighbormap size: " << (int)neighborMap.size() << endl;
	Debug << "mergequeue: " << (int)mMergeQueue.size() << endl;
	Debug << "leaves in queue: " << numLeaves << endl;
	Debug << "overall cost: " << BspMergeCandidate::sOverallCost << endl;

	CLEAR_CONTAINER(bspRays);

	//-- collect the leaves which haven't been found by ray casting
	if (0)
	{
		vector<BspLeaf *> leaves;
		CollectLeaves(leaves, true, mMaxPvs);
		Debug << "found " << (int)leaves.size() << " new leaves" << endl << endl;
		CollectMergeCandidates(leaves);
	}

	return numLeaves;
}


void VspBspTree::ConstructBspRays(vector<BspRay *> &bspRays,
								  const VssRayContainer &rays)
{
	VssRayContainer::const_iterator it, it_end = rays.end();

	for (it = rays.begin(); it != rays.end(); ++ it)
	{
		VssRay *vssRay = *it;
		BspRay *ray = new BspRay(vssRay);

		ViewCellContainer viewCells;

		Ray hray(*vssRay);
		float tmin = 0, tmax = 1.0;
		// matt TODO: remove this!!
		//hray.Init(ray.GetOrigin(), ray.GetDir(), Ray::LINE_SEGMENT);
		if (!mBox.GetRaySegment(hray, tmin, tmax) || (tmin > tmax))
			continue;

		Vector3 origin = hray.Extrap(tmin);
		Vector3 termination = hray.Extrap(tmax);
	
		// cast line segment to get intersections with bsp leaves
		CastLineSegment(origin, termination, viewCells);

		ViewCellContainer::const_iterator vit, vit_end = viewCells.end();
		for (vit = viewCells.begin(); vit != vit_end; ++ vit)
		{
			BspViewCell *vc = dynamic_cast<BspViewCell *>(*vit);
			vector<BspLeaf *>::const_iterator it, it_end = vc->mLeaves.end();
			//NOTE: not sorted!
			for (it = vc->mLeaves.begin(); it != it_end; ++ it)
				ray->intersections.push_back(BspIntersection(0, *it));
		}

		bspRays.push_back(ray);
	}
}


int VspBspTree::MergeViewCells(const VssRayContainer &rays)
{
	BspMergeCandidate::sMaxPvsSize = mMaxPvs;

	MergeStatistics mergeStats;
	mergeStats.Start();
	// TODO: REMOVE LATER for performance!
	const bool showMergeStats = true;
	//BspMergeCandidate::sOverallCost = mBox.SurfaceArea() * mStat.maxPvs;
	long startTime = GetTime();

	if (mUseRaysForMerge)
		mergeStats.nodes = CollectMergeCandidates(rays);
	else
	{
		vector<BspLeaf *> leaves;
		CollectLeaves(leaves);
		mergeStats.nodes = CollectMergeCandidates(leaves);
	}

	mergeStats.collectTime = TimeDiff(startTime, GetTime());
	mergeStats.candidates = (int)mMergeQueue.size();
	startTime = GetTime();

	int nViewCells = mStat.Leaves() - mStat.invalidLeaves;

	//-- use priority queue to merge leaf pairs
	while (!mMergeQueue.empty() && (nViewCells > mMergeMinViewCells) &&
		   (mMergeQueue.top().GetMergeCost() <
		    mMergeMaxCostRatio * BspMergeCandidate::sOverallCost))
	{
		/*Debug << "abs mergecost: " << mMergeQueue.top().GetMergeCost() << " rel mergecost: "
			  << mMergeQueue.top().GetMergeCost() / BspMergeCandidate::sOverallCost 
			  << " max ratio: " << mMergeMaxCostRatio << endl;*/
		BspMergeCandidate mc = mMergeQueue.top();
		mMergeQueue.pop();

		// both view cells equal!
		if (mc.GetLeaf1()->GetViewCell() == mc.GetLeaf2()->GetViewCell())
			continue;

		if (mc.Valid())
		{
			ViewCell::NewMail();
	
			MergeViewCells(mc.GetLeaf1(), mc.GetLeaf2());
			-- nViewCells;
			
			++ mergeStats.merged;
			
			// increase absolute merge cost
			BspMergeCandidate::sOverallCost += mc.GetMergeCost();

			if (showMergeStats)
			{
				if (mc.GetLeaf1()->IsSibling(mc.GetLeaf2()))
					++ mergeStats.siblings;

				const int dist = 
					TreeDistance(mc.GetLeaf1(), mc.GetLeaf2());
				if (dist > mergeStats.maxTreeDist)
					mergeStats.maxTreeDist = dist;
				mergeStats.accTreeDist += dist;
			}
		}
		// merge candidate not valid, because one of the leaves was already
		// merged with another one => validate and reinsert into queue
		else
		{ 
			mc.SetValid();
			mMergeQueue.push(mc);
		}
	}

	mergeStats.mergeTime = TimeDiff(startTime, GetTime());
	mergeStats.Stop();

	if (showMergeStats)
		Debug << mergeStats << endl << endl;
	
	//TODO: should return sample contributions?
	return mergeStats.merged;
}


ViewCell *VspBspTree::GetViewCell(const Vector3 &point)
{
  if (mRoot == NULL)
	return NULL;
  
  stack<BspNode *> nodeStack;
  nodeStack.push(mRoot);
  
  ViewCell *viewcell = NULL;
  
  while (!nodeStack.empty())  {
	BspNode *node = nodeStack.top();
	nodeStack.pop();
	
	if (node->IsLeaf()) {
	  viewcell = dynamic_cast<BspLeaf *>(node)->GetViewCell();
	  break;
	} else {
	  
	  BspInterior *interior = dynamic_cast<BspInterior *>(node);
		
	  // random decision
	  if (interior->GetPlane().Side(point) < 0)
		nodeStack.push(interior->GetBack());
	  else
		nodeStack.push(interior->GetFront());
	}
  }
  
  return viewcell;
}


int VspBspTree::RefineViewCells(const VssRayContainer &rays)
{
	int shuffled = 0;

	Debug << "refining " << (int)mMergeQueue.size() << " candidates " << endl;
	BspLeaf::NewMail();

	// Use priority queue of remaining leaf pairs 
	// These candidates either share the same view cells or
	// are border leaves which share a boundary.
	// We test if they can be shuffled, i.e.,
	// either one leaf is made part of one view cell or the other
	// leaf is made part of the other view cell. It is tested if the
	// remaining view cells are "better" than the old ones.
	while (!mMergeQueue.empty())
	{
		BspMergeCandidate mc = mMergeQueue.top();
		mMergeQueue.pop();

		// both view cells equal or already shuffled
		if ((mc.GetLeaf1()->GetViewCell() == mc.GetLeaf2()->GetViewCell()) ||
			(mc.GetLeaf1()->Mailed()) || (mc.GetLeaf2()->Mailed()))
			continue;
		
		// candidate for shuffling
		const bool wasShuffled = 
			ShuffleLeaves(mc.GetLeaf1(), mc.GetLeaf2());
		
		//-- stats
		if (wasShuffled)
			++ shuffled;
	}

	return shuffled;
}


inline int AddedPvsSize(ObjectPvs pvs1, const ObjectPvs &pvs2)
{
	return pvs1.AddPvs(pvs2);
}

/*inline int SubtractedPvsSize(BspViewCell *vc, BspLeaf *l, const ObjectPvs &pvs2)
{
	ObjectPvs pvs;
	vector<BspLeaf *>::const_iterator it, it_end = vc->mLeaves.end();
	for (it = vc->mLeaves.begin(); it != vc->mLeaves.end(); ++ it)
		if (*it != l)
			pvs.AddPvs(*(*it)->mPvs);
	return pvs.GetSize();
}*/

inline int SubtractedPvsSize(ObjectPvs pvs1, const ObjectPvs &pvs2)
{
	return pvs1.SubtractPvs(pvs2);
}


float GetShuffledVcCost(BspLeaf *leaf, BspViewCell *vc1, BspViewCell *vc2)
{
	//const int pvs1 = SubtractedPvsSize(vc1, leaf, *leaf->mPvs);
	const int pvs1 = SubtractedPvsSize(vc1->GetPvs(), *leaf->mPvs);
	const int pvs2 = AddedPvsSize(vc2->GetPvs(), *leaf->mPvs);

	const float area1 = vc1->GetArea() - leaf->mArea;
	const float area2 = vc2->GetArea() + leaf->mArea;

	const float cost1 = pvs1 * area1;
	const float cost2 = pvs2 * area2;

	return cost1 + cost2;
}


void VspBspTree::ShuffleLeaf(BspLeaf *leaf, 
							 BspViewCell *vc1, 
							 BspViewCell *vc2) const
{
	// compute new pvs and area
	vc1->GetPvs().SubtractPvs(*leaf->mPvs);
	vc2->GetPvs().AddPvs(*leaf->mPvs);
	
	vc1->SetArea(vc1->GetArea() - leaf->mArea);
	vc2->SetArea(vc2->GetArea() + leaf->mArea);

	/// add to second view cell
	vc2->mLeaves.push_back(leaf);

	// erase leaf from old view cell
	vector<BspLeaf *>::iterator it = vc1->mLeaves.begin();

	for (; *it != leaf; ++ it);
	vc1->mLeaves.erase(it);

	/*vc1->GetPvs().mEntries.clear();
	for (; it != vc1->mLeaves.end(); ++ it)
	{
		if (*it == leaf)
			vc1->mLeaves.erase(it);
		else
			vc1->GetPvs().AddPvs(*(*it)->mPvs);
	}*/

	leaf->SetViewCell(vc2); // finally change view cell
}


bool VspBspTree::ShuffleLeaves(BspLeaf *leaf1, BspLeaf *leaf2) const
{
	BspViewCell *vc1 = leaf1->GetViewCell();
	BspViewCell *vc2 = leaf2->GetViewCell();

	const float cost1 = vc1->GetPvs().GetSize() * vc1->GetArea();
	const float cost2 = vc2->GetPvs().GetSize() * vc2->GetArea();

	const float oldCost = cost1 + cost2;
	
	float shuffledCost1 = Limits::Infinity;
	float shuffledCost2 = Limits::Infinity;

	// the view cell should not be empty after the shuffle
	if (vc1->mLeaves.size() > 1)
		shuffledCost1 = GetShuffledVcCost(leaf1, vc1, vc2);
	if (vc2->mLeaves.size() > 1)
		shuffledCost2 = GetShuffledVcCost(leaf2, vc2, vc1);

	// shuffling unsuccessful
	if ((oldCost <= shuffledCost1) && (oldCost <= shuffledCost2))
		return false;
	
	if (shuffledCost1 < shuffledCost2)
	{
		ShuffleLeaf(leaf1, vc1, vc2);
		leaf1->Mail();
	}
	else
	{
		ShuffleLeaf(leaf2, vc2, vc1);
		leaf2->Mail();
	}

	return true;
}

bool VspBspTree::ViewPointValid(const Vector3 &viewPoint) const
{
	BspNode *node = mRoot;

	while (1)
	{
		// early exit
		if (node->TreeValid())
			return true;

		if (node->IsLeaf())
			return false;
			
		BspInterior *in = dynamic_cast<BspInterior *>(node);
					
		if (in->GetPlane().Side(viewPoint) <= 0) 
		{
			node = in->GetBack();
		} 
		else
		{
			node = in->GetFront();
		}
	}

	// should never come here
	return false;
}


bool VspBspTree::CheckValid(const VspBspTraversalData &data) const
{
	return data.mPvs <= mMaxPvs;
}


void VspBspTree::PropagateUpValidity(BspNode *node)
{
	while (!node->IsRoot() && node->GetParent()->TreeValid())
	{
		node = node->GetParent();
		node->SetTreeValid(false);
	}
}


bool VspBspTree::Export(ofstream &stream)
{
	ExportNode(mRoot, stream);

	return true;
}


void VspBspTree::ExportNode(BspNode *node, ofstream &stream)
{
	if (node->IsLeaf())
	{
		BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);
			
		int id = -1;
		if (leaf->GetViewCell() != mOutOfBoundsCell)
			id = leaf->GetViewCell()->GetId();

		stream << "<Leaf viewCellId=\"" << id << "\" />" << endl;
	}
	else
	{
		BspInterior *interior = dynamic_cast<BspInterior *>(node);
	
		Plane3 plane = interior->GetPlane();
		stream << "<Interior plane=\"" << plane.mNormal.x << " " 
			   << plane.mNormal.y << " " << plane.mNormal.z << " " 
			   << plane.mD << "\">" << endl;

		ExportNode(interior->GetBack(), stream);
		ExportNode(interior->GetFront(), stream);

		stream << "</Interior>" << endl;
	}
}

/************************************************************************/
/*                BspMergeCandidate implementation                      */
/************************************************************************/


BspMergeCandidate::BspMergeCandidate(BspLeaf *l1, BspLeaf *l2):
mMergeCost(0),
mLeaf1(l1),
mLeaf2(l2),
mLeaf1Id(l1->GetViewCell()->mMailbox),
mLeaf2Id(l2->GetViewCell()->mMailbox)
{
	EvalMergeCost();
}

float BspMergeCandidate::GetCost(ViewCell *vc) const
{
	return vc->GetPvs().GetSize() * vc->GetArea();
}

float BspMergeCandidate::GetLeaf1Cost() const
{
	BspViewCell *vc = mLeaf1->GetViewCell();
	return GetCost(vc);
}

float BspMergeCandidate::GetLeaf2Cost() const
{
	BspViewCell *vc = mLeaf2->GetViewCell();
	return GetCost(vc);
}

void BspMergeCandidate::EvalMergeCost()
{
	//-- compute pvs difference
	BspViewCell *vc1 = mLeaf1->GetViewCell();
	BspViewCell *vc2 = mLeaf2->GetViewCell();

	const int diff1 = vc1->GetPvs().Diff(vc2->GetPvs());
	const int newPvs = diff1 + vc1->GetPvs().GetSize();

	//-- compute ratio of old cost
	//   (i.e., added size of left and right view cell times pvs size)
	//   to new rendering cost (i.e, size of merged view cell times pvs size)
	const float oldCost = GetLeaf1Cost() + GetLeaf2Cost();

	const float newCost =
		(float)newPvs * (vc1->GetArea() + vc2->GetArea());

	if (newPvs > sMaxPvsSize) // strong penalty if pvs size too large
	{
		mMergeCost = 1e15f;
	}
	else
	{
		mMergeCost = newCost - oldCost;
	}
}


void BspMergeCandidate::SetLeaf1(BspLeaf *l)
{
	mLeaf1 = l;
}


void BspMergeCandidate::SetLeaf2(BspLeaf *l)
{
	mLeaf2 = l;
}


BspLeaf *BspMergeCandidate::GetLeaf1()
{
	return mLeaf1;
}


BspLeaf *BspMergeCandidate::GetLeaf2()
{
	return mLeaf2;
}


bool BspMergeCandidate::Valid() const
{
	return
		(mLeaf1->GetViewCell()->mMailbox == mLeaf1Id) &&
		(mLeaf2->GetViewCell()->mMailbox == mLeaf2Id);
}


float BspMergeCandidate::GetMergeCost() const
{
	return mMergeCost;
}


void BspMergeCandidate::SetValid()
{
	mLeaf1Id = mLeaf1->GetViewCell()->mMailbox;
	mLeaf2Id = mLeaf2->GetViewCell()->mMailbox;

	EvalMergeCost();
}


/************************************************************************/
/*                  MergeStatistics implementation                      */
/************************************************************************/


void MergeStatistics::Print(ostream &app) const
{
	app << "===== Merge statistics ===============\n";

	app << setprecision(4);

	app << "#N_CTIME ( Overall time [s] )\n" << Time() << " \n";

	app << "#N_CCTIME ( Collect candidates time [s] )\n" << collectTime * 1e-3f << " \n";

	app << "#N_MTIME ( Merge time [s] )\n" << mergeTime * 1e-3f << " \n";

	app << "#N_NODES ( Number of nodes before merge )\n" << nodes << "\n";

	app << "#N_CANDIDATES ( Number of merge candidates )\n" << candidates << "\n";

	app << "#N_MERGEDSIBLINGS ( Number of merged siblings )\n" << siblings << "\n";
	app << "#N_MERGEDNODES ( Number of merged nodes )\n" << merged << "\n";

	app << "#MAX_TREEDIST ( Maximal distance in tree of merged leaves )\n" << maxTreeDist << "\n";

	app << "#AVG_TREEDIST ( Average distance in tree of merged leaves )\n" << AvgTreeDist() << "\n";

	app << "===== END OF BspTree statistics ==========\n";
}
